#include "mdp.h"
#include "infer_old.h"

#define DEBUGEM(x) //x
#ifndef rescaleRewards
#  define rescaleRewards 0
#endif

void glDisplayGrey   (const arr &x,uint d0,uint d1,bool wait,uint win);
void glDisplayRedBlue(const arr &x,uint d0,uint d1,bool wait,uint win);

#undef Pt
#define Pt(t) ::pow(mdp.gamma,t)

double EMstep_lev1_align(
      const MDP& mdp,
      FSC_lev1& fsc,
      uint T,
      bool stepDebug,
      bool maxMstep){
  
  CHECK(mdp.Px.nd==1 && mdp.Pxax.nd==3 &&
        mdp.Pyxa.nd==3 && mdp.Rax.nd==2 &&
        fsc.Pa0.nd==2 && fsc.P0y0.nd==3,"");
  
  uint
    dx=mdp.Pxax.d0,
    da=mdp.Pxax.d1,	
    dy=mdp.Pyxa.d0,
    d0=fsc.P0y0.d0;

  uint t;

  MaxMarginalMode=false;

#if rescaleRewards
  double Rmin=Rax.min(),Rmax=Rax.max();
  for(uint i=0;i<Rax.N;i++) Rax.elem(i) = (Rax.elem(i)-Rmin)/(Rmax-Rmin);
#else
  double Rmin=0.,Rmax=1.;
#endif

  //step 0: eliminate a
  arr Px0x(dx,d0,dx), R0x(d0,dx), Pyx0(dy,dx,d0);
  tensor(Px0x, mdp.Pxax, TUP(0,3,2), fsc.Pa0, TUP(3,1), 1);
  tensor(Pyx0, mdp.Pyxa, TUP(0,1,3), fsc.Pa0, TUP(3,2), 1);
  tensor(R0x , mdp.Rax , TUP(2,1)  , fsc.Pa0, TUP(2,0), 1);

  //step 1: eliminate y:
  arr P0x0(d0,dx,d0);
  tensor(P0x0, fsc.P0y0, TUP(0,3,2), Pyx0, TUP(3,1,2), 1);

  DEBUGEM(
		Px0x.tensorCheckCondNormalization(1);
	  Pyx0.tensorCheckCondNormalization(1); 
	  P0x0.tensorCheckCondNormalization(1); );
	
  //step 2: create 2-time-slice factor graph
  uint x,y,n0,a,x_,y_,n0_,a_; //id's for the variables
  FactorModel M;
  //variables
  x   = M.variable(dx ,"state(t)");
  y   = M.variable(dy ,"observation(t)");
  n0  = M.variable(d0 ,"node0(t)");
  a   = M.variable(da ,"action(t)");
  x_  = M.variable(dx ,"state(t+1)");
  y_  = M.variable(dy ,"observation(t+1)");
  n0_ = M.variable(d0 ,"node0(t+1)");
  a_  = M.variable(da ,"action(t+1)");
  //factors
	TableFac tmp(M.V,TUP(x));
  M.factor(TableFT,TUP(x_,n0,x)  ,Px0x);
  M.factor(TableFT,TUP(n0_,x_,n0),P0x0);
	
  //edges
  M.arcs(TUP(0,1)); //(all edges implicitly labeled `0')
  allocateEdgeFactors(M);
  storeOrgFactors(M);

  //step 3: make 2 copys of the factor graph for fwd and bwd iteration...
  FactorModel Mfwd,Mbwd;
  Mbwd=Mfwd=M;

  //step 4: initialize fwd and bwd models:
  Mfwd(0)->multiply(TableFac(M.V,TUP(x )  ,mdp.Px));
  Mfwd(0)->multiply(TableFac(M.V,TUP(n0)  ,fsc.P0));
  Mbwd(0)->multiply(TableFac(M.V,TUP(n0,x),R0x));
  DEBUGEM(
    storeOrgFactors(Mfwd);
    storeOrgFactors(Mbwd);
    checkConsistency(Mfwd);
    checkConsistency(Mbwd); );

  //-- prepare some buffers for fwd-bwd
  //buffers for closing the loop
  TableFac head  (M.V,TUP(n0 ,x ));
  TableFac tail  (M.V,TUP(n0 ,x ));
  Mfwd(0)->marginal(head);
  Mbwd(0)->marginal(tail);
  
  //overlab measures
  arr L(2*T); L.setZero();
  L(0) = sum(head.P % tail.P)*::exp(head.logP+tail.logP);
  arr La(T),Lb(T); La.setZero(); Lb.setZero();

  //collectors
  TableFac alpha(M.V,TUP(n0 ,x )); alpha.P=0.; alpha.logP=0.;
  Mfwd(0)->marginal(alpha);
  TableFac beta (M.V,TUP(n0_,x_)); beta .P=0.; beta .logP=0.;
  //Mbwd(1)->marginal(beta);

  //marginals for display
  TableFac xmarg(M.V,TUP(x)),nmarg(M.V,TUP(n0));


  //***********!!!!!!!!!!!********** switch to max product?
  MaxMarginalMode=false;

  //step 5: iterate
  MT::resetTimer();
  for(t=0;t<T;t++){
    //display some marginal
    if(stepDebug){
      //display head's x-marginal
			cout <<"t=" <<t;
      xmarg.id=x;
      head.marginal(xmarg);
      glDisplayGrey(xmarg.P*::exp(xmarg.logP),global_maze.d0,global_maze.d1,false,0);
			cout <<" alpha_x_norm=" <<::exp(xmarg.logP);
      La(t) = xmarg.P(xmarg.P.N-1)*::exp(xmarg.logP);
      //display head's n-marginal
      nmarg.id=n0;
			head.marginal(nmarg);
      glDisplayGrey(nmarg.P,1,d0,false,2);
			cout <<" alpha_n_norm=" <<::exp(nmarg.logP);
      //display tail's x-marginal
      xmarg.id=x_;
      tail.marginal(xmarg);
      glDisplayRedBlue(xmarg.P*::exp(xmarg.logP),global_maze.d0,global_maze.d1,false,1);
			cout <<" beta_x_norm=" <<::exp(xmarg.logP);
      Lb(t) = xmarg.P(0)*::exp(xmarg.logP);
			cout <<endl;
			MT::wait();
    }

    //propagate fwd and bwd
#if 1 //expensive way
    passLabeledEdges(Mfwd,0,FORWARD);
    passLabeledEdges(Mbwd,0,BACKWARD);
#else //maybe cheaper way
    Mfwd(0)->marginal(nu);
    Mfwd(1)->multiply(nu);
    Mbwd(1)->marginal(nu);
    Mbwd(0)->multiply(nu);
#endif
    DEBUGEM( passLabeledEdges(Mbwd,0,FORWARD); //don't need this really...
	     checkConsistency(Mfwd);
	     checkEquilibrium(Mfwd);
	     checkConsistency(Mbwd);
	     checkEquilibrium(Mbwd); );

    //get tail from backward model
    tail.id = TUP(n0 ,x );
    Mbwd(0)->marginal(tail);

    //likelihood for T=t+(t-1), before updating head
    L(t+t)   = sum(head.P % tail.P)*::exp(head.logP+tail.logP);

    //get head from the forward model
    head.id = TUP(n0_,x_);
    Mfwd(1)->marginal(head);

    //likelihood for T=t+t
    L(t+t+1) = sum(head.P % tail.P)*::exp(head.logP+tail.logP);
    DEBUGEM( cout <<"head log=" <<head.logP <<" tail log=" <<tail.logP <<endl;  );

    //accumulation of alpha and beta
    head.id = TUP(n0 ,x );
    tail.id = TUP(n0_,x_);
    alpha.weightedAdd(Pt(t+1),head);
    beta .weightedAdd(Pt(t+1),tail);

    //reset fwd and bwd models for next iteration
    Mfwd=M;
    Mbwd=M;

    //set head and tail
    Mfwd(0)->multiply(head);
    Mbwd(1)->multiply(tail);
    DEBUGEM( storeOrgFactors(Mfwd);
	     storeOrgFactors(Mbwd);
	     checkConsistency(Mfwd);
	     checkConsistency(Mbwd); );

  }
  cout <<"E: " <<MT::getTimer() <<"sec, M: " <<std::flush;

  DEBUGEM(
		 xmarg.id=x;
		 alpha.marginal(xmarg);
	   glDisplayGrey(xmarg.P,global_maze.d0,global_maze.d1);
	   xmarg.id=x_;
	   beta.marginal(xmarg);
	   glDisplayRedBlue(xmarg.P,global_maze.d0,global_maze.d1);
	   //MT::wait();
		 );

  DEBUGEM(
		cout <<"alpha=" <<alpha <<endl;
	  //MT::wait();
	  cout <<"beta=" <<beta <<endl;
	  //MT::wait();
		);

  //MT::save(beta,"z.beta");
  //if(beta.P.min()<0.) beta.P+=beta.P.min();
  //beta.logP=0;
  //MT::save(beta,"z.beta2");
  //beta.P.tensorCondSoftMax(2,20.);

  //***********!!!!!!!!!!!********** switch to max product
  //MaxMarginalMode=true;

  //step 6: setup detailed (including y and a) factor graph over 2-time-slice
  M.clearFactors();
  //factors
  M.factor(TableFT,TUP(n0_,y_,x_,a,n0,x));

  //step 7: seed factors with alpha and beta
  //potentials on clique 1: alpha Pyx0 P0y0
  M(0)->multiply(TableFac(M.V,TUP( a ,n0   ),fsc.Pa0));
  M(0)->multiply(TableFac(M.V,TUP( x_,a ,x ),mdp.Pxax));
  M(0)->multiply(TableFac(M.V,TUP( y_,x_,a ),mdp.Pyxa));
  M(0)->multiply(TableFac(M.V,TUP(n0_,y_,n0),fsc.P0y0));
#if 1
  M(0)->multiply(alpha);
#else //also mix the alpha_{t=0} into the alpha (already done in the initialization of alpha)
  TableFac alphaFac(M.V,TUP(n0,x));
  alphaFac.multiply(alpha);
  TableFac startFac(M.V,TUP(n0,x));
  startFac.multiply(TableFac(M.V,TUP(x ),mpd.Px));
  startFac.multiply(TableFac(M.V,TUP(n0),fsc.P0));
  startFac.weightedAdd(1.,alphaFac);
  M(0)->multiply(startFac);
#endif

  //potentials on clique 2: beta Pa0 Pxax
#if 0
  M(1)->multiply(beta);
#else //also mix the beta_{t=0} into the beta
  TableFac betaFac  (M.V,TUP(n0_,x_,a,x));
  betaFac.multiply(beta);
  TableFac rewardFac(M.V,TUP(n0_,x_,a,x));
  rewardFac.multiply(TableFac(M.V,TUP(a,x),mdp.Rax));
  rewardFac.weightedAdd(1.,betaFac);
  M(0)->multiply(rewardFac);
#endif

  DEBUGEM( storeOrgFactors(M);
	   checkConsistency(M); );

  //step 8: inference on the full 2-time-slice
  passLabeledEdges(M,0,FORWARD);
  passLabeledEdges(M,0,BACKWARD);
  DEBUGEM( checkConsistency(M);
	   checkEquilibrium(M); );

  // M-step: extract posterior marginals
	TableFac X0y0(M.V,TUP(n0_,y_,n0));
	TableFac Xa0 (M.V,TUP(a ,n0 ));
  M(0)->marginal(Xa0);
  M(0)->marginal(X0y0);

  X0y0.P.reshape(d0,dy,d0);
#if !maxMstep
  fsc.P0y0 = X0y0.P;
#else
  X0y0.P /= fsc.P0y0;
  X0y0.P.tensorCondMax(1);
  X0y0.P += 3.;
  rndUni(X0y0.P,.0,1e-3,true);
	rndUni(fsc.P0y0,.0001,.00011,true);
  fsc.P0y0 *= X0y0.P;
#endif
  fsc.P0y0.tensorCondNormalize(1);

  Xa0.P.reshape(da,d0);
#if !maxMstep
  fsc.Pa0 = Xa0.P;
#else
  Xa0.P /= fsc.Pa0;
  Xa0.P.tensorCondMax(1);
  Xa0.P += 3.;
  rndUni(Xa0.P,.0,1e-3,true);
	rndUni(fsc.Pa0,.0001,.00011,true);
  fsc.Pa0 *= Xa0.P;
#endif
  fsc.Pa0.tensorCondNormalize(1);

  fsc.Pa0.tensorCheckCondNormalization(1);
  fsc.P0y0.tensorCheckCondNormalization(1);

  //----- rest is cosmetics
  //report
  cout <<MT::getTimer() <<"sec," <<std::flush;
  double Ltot=0.,PtNorm=0.,Texp=0.;
  arr PtL(2*T);
  for(t=0;t<2*T;t++){ PtL(t)=Pt(t)*L(t); Ltot+=PtL(t); Texp+=t*PtL(t); PtNorm+=Pt(t); }
  PtL  /= PtNorm;
  Texp /= Ltot;
  Ltot /= PtNorm; //(in case Pt was not normalized)
  cout <<" P(r=1)=" <<Ltot
       <<", Exp(T)=" <<Texp <<"/"<<::log(Ltot*PtNorm)/::log(mdp.gamma)
       <<", Exp(R)=" <<(Ltot*(Rmax-Rmin)+Rmin)*PtNorm
       <<endl;

  //displays
  if(true || stepDebug){
    glDisplayGrey(fsc.P0y0,d0*dy,d0,false,2);
    glDisplayGrey(fsc.Pa0 ,da,d0,false,1);
    L.resizeCopy(T); //make shorter for storage...
    write(L,PtL,La,Lb,"z.L");
    //gnuplot("plot 'z.L' us 1 title 'L(t)','z.L' us 2 title 'La(t)' with points,'z.L' us 3 title 'Lb(t)' with points");
  }
  
  return (Ltot*(Rmax-Rmin)+Rmin)*PtNorm;
}
